import asyncio
from typing import Any, AsyncIterator, Callable, Coroutine, Set

from nextline import Nextline
from nextline.utils import pubsub

from nextline_schedule.fsm import build_state_machine


class AutoMode:
    '''A model of the finite state machine.

    >>> nextline = Nextline(statement='')
    >>> async def request_statement() -> str:
    ...     return ''

    >>> auto_mode = AutoMode(nextline=nextline, request_statement=request_statement)
    >>> auto_mode.state
    'created'
    '''

    def __init__(
        self,
        nextline: Nextline,
        request_statement: Callable[[], Coroutine[Any, Any, str]],
    ):
        self._nextline = nextline
        self._request_statement = request_statement
        self._tasks: Set[asyncio.Task] = set()
        self._pubsub_state = pubsub.PubSubItem[str]()

        machine = build_state_machine(self)
        machine.after_state_change = self.after_state_change.__name__  # type: ignore

    def subscribe_state(self) -> AsyncIterator[str]:
        return self._pubsub_state.subscribe()

    async def on_enter_waiting(self) -> None:
        async def wait() -> None:
            async for state in self._nextline.subscribe_state():
                if state == 'initialized':
                    await self.on_initialized()  # type: ignore
                    break
                if state == 'finished':
                    await self.on_finished()  # type: ignore
                    break
            return

        task = asyncio.create_task(wait())
        self._tasks.add(task)

    async def on_enter_auto_pulling(self) -> None:
        async def pull() -> None:
            statement = await self._request_statement()
            await self._nextline.reset(statement=statement)
            await self.run()  # type: ignore

        task = asyncio.create_task(pull())
        self._tasks.add(task)

    async def on_enter_auto_running(self) -> None:
        async def run() -> None:
            async def run_and_wait() -> None:
                async with self._nextline.run_session():
                    pass
                if self._nextline.exception() is not None:
                    await self.on_raised()  # type: ignore
                else:
                    await self.on_finished()  # type: ignore

            async def continue_on_prompt(nextline: Nextline) -> None:
                async for prompt_info in nextline.subscribe_prompt_info():
                    if prompt_info.trace_call_end:  # TODO: remove when unnecessary
                        continue
                    if not prompt_info.open:
                        continue
                    break
                await nextline.send_pdb_command(
                    command='continue',
                    prompt_no=prompt_info.prompt_no,
                    trace_no=prompt_info.trace_no,
                )

            await asyncio.gather(
                run_and_wait(),
                continue_on_prompt(self._nextline),
                # monitor(),
            )

        task = asyncio.create_task(run())
        self._tasks.add(task)

    async def after_state_change(self) -> None:
        await self._collect_tasks()
        await self._pubsub_state.publish(self.state)  # type: ignore

    async def _collect_tasks(self) -> None:
        if not self._tasks:
            return
        done, _ = await asyncio.wait(
            self._tasks, timeout=0, return_when=asyncio.FIRST_COMPLETED
        )
        self._tasks -= done

    async def close(self) -> None:
        await self._collect_tasks()
        await self._pubsub_state.close()

    async def __aenter__(self) -> 'AutoMode':
        await self.start()  # type: ignore
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback
        await self.close()



import strawberry
from strawberry.types import Info


def query_auto_mode_state(info: Info) -> str:
    auto_mode = info.context["auto_mode"]
    return auto_mode.state


@strawberry.type
class QueryAutoMode:
    state: str = strawberry.field(resolver=query_auto_mode_state)


def query_scheduler_api_url(info: Info) -> str:
    scheduler = info.context["scheduler"]
    return scheduler._api_url


def query_scheduler_length_minutes(info: Info) -> int:
    scheduler = info.context["scheduler"]
    return scheduler._length_minutes


def query_scheduler_policy(info: Info) -> str:
    scheduler = info.context["scheduler"]
    return scheduler._policy


@strawberry.type
class QueryScheduler:
    api_url: str = strawberry.field(resolver=query_scheduler_api_url)
    length_minutes: int = strawberry.field(resolver=query_scheduler_length_minutes)
    policy: str = strawberry.field(resolver=query_scheduler_policy)


@strawberry.type
class QuerySchedule:
    @strawberry.field
    def auto_mode(self, info: Info) -> QueryAutoMode:
        return QueryAutoMode()

    @strawberry.field
    def scheduler(self, info: Info) -> QueryScheduler:
        return QueryScheduler()


@strawberry.type
class Query:
    @strawberry.field
    def schedule(self, info: Info) -> QuerySchedule:
        return QuerySchedule()



from typing import Optional

import strawberry
from strawberry.types import Info


async def mutate_turn_on(info: Info) -> bool:
    auto_mode = info.context["auto_mode"]
    await auto_mode.turn_on()
    return True


async def mutate_turn_off(info: Info) -> bool:
    auto_mode = info.context["auto_mode"]
    await auto_mode.turn_off()
    return True


@strawberry.type
class MutationAutoMode:
    turn_on: bool = strawberry.field(resolver=mutate_turn_on)
    turn_off: bool = strawberry.field(resolver=mutate_turn_off)


@strawberry.input
class MutationSchedulerInput:
    api_url: Optional[str] = None
    length_minutes: Optional[int] = None
    policy: Optional[str] = None


@strawberry.type
class MutationScheduler:
    @strawberry.mutation
    def update(self, info: Info, input: MutationSchedulerInput) -> bool:
        scheduler = info.context["scheduler"]
        if input.api_url is not None:
            scheduler._api_url = input.api_url
        if input.length_minutes is not None:
            scheduler._length_minutes = input.length_minutes
        if input.policy is not None:
            scheduler._policy = input.policy
        return True


@strawberry.type
class MutationSchedule:
    @strawberry.field
    def auto_mode(self, info: Info) -> MutationAutoMode:
        return MutationAutoMode()

    @strawberry.field
    def scheduler(self, info: Info) -> MutationScheduler:
        return MutationScheduler()


@strawberry.type
class Mutation:
    @strawberry.field
    def schedule(self, info: Info) -> MutationSchedule:
        return MutationSchedule()



from copy import deepcopy

import pytest
from transitions import Machine
from transitions.extensions.markup import HierarchicalMarkupMachine

from nextline_schedule.fsm import build_state_machine


def test_model_default():
    machine = build_state_machine()
    assert not machine.models
    print(machine.models)


def test_model_self_literal():
    machine = build_state_machine(model=Machine.self_literal)
    assert machine.models[0] is machine
    assert len(machine.models) == 1


async def test_transitions() -> None:
    machine = build_state_machine(model=Machine.self_literal)
    await machine.start()

    # off -- turn_on --> waiting -- turn_off --> off
    assert machine.is_off()
    await machine.turn_on()
    assert machine.is_waiting()
    await machine.turn_off()

    # off -- turn_on --> waiting -- on_initialized --> auto_pulling -- turn_off --> off
    assert machine.is_off()
    await machine.turn_on()
    await machine.on_initialized()
    assert machine.is_auto_pulling()
    await machine.turn_off()

    # off -- turn_on --> waiting -- on_finished --> auto_pulling -- turn_off --> off
    assert machine.is_off()
    await machine.turn_on()
    await machine.on_finished()
    assert machine.is_auto_pulling()
    await machine.turn_off()

    # off -- turn_on --> waiting -- on_finished --> auto_pulling -- run --> auto_running -- turn_off --> off
    assert machine.is_off()
    await machine.turn_on()
    await machine.on_finished()
    await machine.run()
    assert machine.is_auto_running()
    await machine.turn_off()

    # off -- turn_on --> waiting -- on_finished --> auto_pulling -- run --> auto_running -- on_raised --> off
    assert machine.is_off()
    await machine.turn_on()
    await machine.on_finished()
    await machine.run()
    await machine.on_finished()
    assert machine.is_auto_pulling()
    await machine.run()
    assert machine.is_auto_running()
    await machine.on_raised()

    assert machine.is_off()


async def test_ignore_invalid_triggers() -> None:
    machine = build_state_machine(model=Machine.self_literal)
    await machine.start()
    assert machine.is_off()
    await machine.on_finished()  # invalid trigger

    assert machine.is_off()
    await machine.turn_on()
    await machine.on_initialized()
    assert machine.is_auto_pulling()
    await machine.on_finished()  # invalid trigger
    assert machine.is_auto_pulling()


def test_restore_from_markup():
    machine = build_state_machine(markup=True)
    assert isinstance(machine.markup, dict)
    markup = deepcopy(machine.markup)
    del markup['models']
    rebuild = HierarchicalMarkupMachine(model=None, **markup)
    assert rebuild.markup == machine.markup


@pytest.mark.skip
def test_graph():
    machine = build_state_machine(model=Machine.self_literal, graph=True)
    machine.get_graph().draw('states.png', prog='dot')



from typing import Type

from transitions import Machine
from transitions.extensions import MachineFactory
from transitions.extensions.markup import HierarchicalMarkupMachine


def build_state_machine(model=None, graph=False, asyncio=True, markup=False) -> Machine:
    '''Finite state machine for the auto-pilot mode states.

    State Diagram:

                 .-------------.
                 |   Created   |
                 '-------------'
                        | start()
                        v
                 .-------------.
            .--->|     Off     |<--------------.
            |    '-------------'               |
            |           | turn_on()            |
         turn_off()     |                      |
            |           v                      |
            |    .-------------.               |
            |----|   Waiting   |               |
            |    '-------------'          on_raised()
            |           |                      |
            |           | on_initialized()     |
            |           | on_finished()        |
        .---------------+------------------.   |
        |   Auto        |                  |   |
        |               v                  |   |
        |        .-------------.           |   |
        |        |   Pulling   |<--.       |   |
        |        '-------------'   |       |   |
        |         run() |          |       |   |
        |               |    on_finished() |   |
        |               v          |       |   |
        |        .-------------.   |       |   |
        |        |   Running   |-----------+---'
        |        '-------------'           |
        |                                  |
        '----------------------------------'

    >>> class Model:
    ...     def on_enter_waiting(self):
    ...         print('enter the waiting state')
    ...         self.on_finished()
    ...
    ...     def on_exit_waiting(self):
    ...        print('exit the waiting state')
    ...
    ...     def on_enter_auto_pulling(self):
    ...        print('enter the pulling state')

    >>> model = Model()
    >>> machine = build_state_machine(model=model, asyncio=False)
    >>> model.state
    'created'

    >>> _ = model.start()
    >>> model.state
    'off'

    >>> _ = model.turn_on()
    enter the waiting state
    exit the waiting state
    enter the pulling state

    >>> model.state
    'auto_pulling'


    '''

    MachineClass: Type[Machine]
    if markup:
        MachineClass = HierarchicalMarkupMachine
    else:
        MachineClass = MachineFactory.get_predefined(
            graph=graph, nested=True, asyncio=asyncio
        )

    # Build a state machine with nested states from dict configs with "remap"
    # in the way described in the docs:
    # https://github.com/pytransitions/transitions/tree/0.9.0#reuse-of-previously-created-hsms

    auto_state_conf = {
        'name': 'auto',
        'states': ['pulling', 'running', 'raised'],
        'transitions': [
            ['run', 'pulling', 'running'],
            ['on_finished', 'running', 'pulling'],
            ['on_raised', 'running', 'raised'],
        ],
        'initial': 'pulling',
        'queued': True,
        'ignore_invalid_triggers': True,
    }

    # Ideally, we would be able to pass the auto_state_conf dict directly to
    # the MachineClass constructor, but this doesn't work with "remap."
    # Instead, we have to create a new instance of the MachineClass and pass it
    # to the parent state machine.
    auto_state = MachineClass(model=None, **auto_state_conf)  # type: ignore

    state_conf = {
        'name': 'global',
        'states': [
            'created',
            'off',
            'waiting',
            {'name': 'auto', 'children': auto_state, 'remap': {'raised': 'off'}},
        ],
        'transitions': [
            ['start', 'created', 'off'],
            ['turn_on', 'off', 'waiting'],
            ['on_initialized', 'waiting', 'auto'],
            ['on_finished', 'waiting', 'auto'],
            ['turn_off', 'waiting', 'off'],
            ['turn_off', 'auto', 'off'],
        ],
        'initial': 'created',
        'queued': True,
        'ignore_invalid_triggers': True,
    }

    machine = MachineClass(model=model, **state_conf)  # type: ignore
    return machine



import random
from datetime import datetime, timedelta

import httpx


class RequestStatement:
    def __init__(self, api_url: str, length_minutes: int, policy: str):
        self._api_url = api_url
        self._length_minutes = length_minutes
        self._policy = policy

    async def __call__(self) -> str:
        # https://github.com/simonsobs/so-scheduler/blob/master/readme.md#schedule-api
        start_time = datetime.utcnow()
        duration = timedelta(minutes=self._length_minutes)
        end_time = start_time + duration
        start_time_str = start_time.strftime('%Y-%m-%d %H:%M')
        end_time_str = end_time.strftime('%Y-%m-%d %H:%M')
        data = {"t0": start_time_str, "t1": end_time_str, "policy": self._policy}
        print(data)
        async with httpx.AsyncClient() as client:
            response = await client.post(self._api_url, json=data)
        print(response.json())
        return response.json()['commands']


class DummyRequestStatement(RequestStatement):
    async def __call__(self) -> str:
        return generate_dummy_statement()


def generate_dummy_statement() -> str:
    s = '\n'.join(
        f'time.sleep({random.uniform(1, 3):.2})' for _ in range(random.randint(5, 10))
    )
    s = '\n'.join(
        (
            f'# {datetime.utcnow().isoformat()}',
            'import time',
            '',
            s,
        )
    )
    return s
